# SPDX-FileCopyrightText: 2022-present Sergej Levich <sergej.levich@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

"""doc2data - Integrated document processing with machine learning."""

__version__ = "0.2.0"
